//
//  RiiiverSDKManager.swift
//  sample
//
//  Created by CITIZEN Watch Co.,Ltd on 2019/10/28.
//  Copyright © 2019 CITIZEN Watch Co.,Ltd. All rights reserved.
//

import Foundation
import WebKit

import RiiiverSDK

/// The class for managing Riiiver SDK information.  -  RiiiverSDKから取得した情報を管理
class RiiiverSDKManager : ERSDKDelegate {
    
    /// Singleton
    static let shared = RiiiverSDKManager()
    
    /// accecss Token  -  認証トークン
    public private(set) var accessToken: String?
    
    /// ID Token
    public private(set) var idToken: String?
    
    private init() {
        ERSDK.sharedInstance.delegate = self
    }
    
    /// 認証済みかどうか (true:認証済み、false:未認証)
    /// Is this user authenticated?  true:  Yes,  false: No
    public var authenticated: Bool {
        return ERAuth.sharedInstance.isLoggedIn()
    }
    
    /// RiiiverSDKからWeb認証画面表示する
    ///  set a web view to be  invoked by Riiiver SDK for the web page authentication.
    /// - Parameters:
    ///   - completion:
    func startWebLogin(completion:((URL?, Error?)->Void)?) {
        let deviceId:[String] = AppManager.shared.getDeviceId()
        // Display the authentication web page.  -  Web認証画面表示
        ERAuth.sharedInstance.startWebLogin(deviceIds: deviceId, completion: completion)
    }
    
    /// RiiiverSDKから認証トークンを取得して保持する
    ///  Get the auth token and cache it.
    /// - Parameters:
    ///   - completion:
    ///     - isSuccess: true > Success to get the authentication info.  false > Failed to get the auth info.  認証情報の取得が成功した場合はtrue、失敗した場合はfalse
    ///     - error:  the error information from Riiiver SDK.  -   RiiiverSDKから返却されたerror情報
    func getAWSToken(completion: @escaping (( _ isSuccess: Bool, _ error: Error?) -> Void)) {
        // Check for the state of the authentication inside of RiiiverSDK.  -  RiiiverSDKの認証状態をチェック
        if (authenticated == false ){
            //  When failed to authenticate, quit this because we couldn't get auth token.  -  未認証の場合、認証トークンを取得できないので終了
            print("getAWSToken failed. Unauthenticated.")
            completion(false, nil)
            return
        }
        
        // RiiiverSDKから認証トークンを取得 - Get the token from Riiiver SDK.
        ERAuth.sharedInstance.getToken {(error: Error?, token: ERAuthToken?) -> Void in
            
            guard error == nil else {
                print("getAWSToken failed. reason:\(error?.localizedDescription ?? "")")
                completion(false, error)
                return
            }
            
            guard let token = token else {
                print("getAWSToken failed. reason: No token")
                completion(false, error)
                return
            }
            
            guard let accessToken = token.accessToken else {
                print("getAWSToken failed. reason: No access-token")
                completion(false, error)
                return
            }
            // 取得した情報を保持 - store the token into app side.   (But never use them, in this app.)
            self.accessToken = token.accessToken
            self.idToken = token.idToken
            
            print("getAWSToken succeeded. accessToken: \(accessToken)")
            completion(true, error)
        }
    }
    
    /// RiiiverSDKからiiideaStore画面表示する
    ///  Display Riiiver Store page via Riiiver SDK.
    /// - Parameters:
    ///   - webView
    ///   - delegate
    func openStore(webView: WKWebView, delegate: ERAuthStoreWebViewDelegate) {
        let deviceId:[String] = AppManager.shared.getDeviceId()
        ERAuth.sharedInstance.openStore(webView: webView, deviceIds: deviceId, delegate: delegate)
    }
    
    /// RiiiverSDKにトークン情報を設定する
    ///
    /// - Parameters:
    ///   - webView
    ///   - delegate
    public func challengeAuthByWebLogin(completion:((Error?,ERAuthToken?)->Void)?) {
        let deviceId:[String] = AppManager.shared.getDeviceId()
        // RiiiverSDKにトークン情報を設定する
        ERAuth.sharedInstance.challengeAuthByWebLogin(deviceIds: deviceId, completion: completion)
    }

    /// ログインユーザーのプロフィール情報をwebViewに表示する
    ///
    /// - Parameters:
    ///   - webView
    ///   - delegate
    public func openProfile(webView: WKWebView, delegate:ERAuthProfileWebViewDelegate) {
        let deviceId:[String] = AppManager.shared.getDeviceId()
        // RiiiverSDKにトークン情報を設定する
        ERAuth.sharedInstance.openProfile(webView: webView, deviceIds: deviceId, delegate: delegate)
    }
    

    /// RiiiverSDKから My iiidea 一覧を取得する
    /// Get the my iiidea list from RiiiverSDK.
    /// - Parameters:
    ///   - completion:
    ///     - isSuccess: My iiidea 一覧を取得が成功した場合はtrue、失敗した場合はfalse
    ///     - error: RiiiverSDKから返却されたerror情報
    func getMyApplets(completion: @escaping ((_ isSuccess: Bool, _ iiideas: [ERApplet]? , _ error: Error?) -> Void)) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        self.getAWSToken { (isSuccess, error) in
            if (isSuccess == false) {
                print("getMyApplets failed. reason: No token")
                completion(false, nil, error)
                return
            }
            // RiiiverSDKから My iiidea 一覧を取得 - get my iiideas from Riiiver SDK.
            ERSDK.sharedInstance.getMyApplets(deviceIds: AppManager.shared.getDeviceId()) { (myApplets, error) in
                
                // Note: エラーがあっても取得できた My iiidea は返却されるため処理を継続することは可能 - Here, we can get the iiidea even if we get some errors.
                if let error = error {
                    print("getMyApplets failed. reason: error:\(error)")
                }

                print("getMyApplets succeeded. My iiidea(\(myApplets?.count ?? 0))")
                completion(true, myApplets, error)
            }
        }
    }
    
    /// 指定したiiideaをアクティベートする
    /// Activate an terget iiidea
    /// - Parameters:
    ///   - iiidea: アクティベートするiiidea
    ///   - completion:
    ///     - isSuccess:アクティベートが成功した場合はtrue、失敗した場合はfalse
    ///     - error: RiiiverSDKから返却されたerror情報
    func activate(iiidea: ERApplet, completion: @escaping ((_ isSuccess: Bool, _ error: Error?) -> Void)) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        self.getAWSToken { (isSuccess, error) in
            if (isSuccess == false) {
                print("activate failed. reason: No token")
                completion(false, error)
                return
            }
            // Activate the iiidea via Riiiver SDK.
            iiidea.activate() { unauthPermissions, serviceProxyData, error in
                guard unauthPermissions == nil || unauthPermissions == [] else {
                    print("activate failed. reason:No permission")
                    completion(false, error)
                    return
                }
                guard error == nil else {
                    print("activate failed. reason: \(error.debugDescription)")
                    completion(false, error)
                    return
                }
                print("activate succeeded.")
                completion(true, error)
            }
        }
    }
    
    /// 指定したiiideaをディアクティベートする
    ///  Deactivate the target iiidea.
    /// - Parameters:
    ///   - iiidea: ディアクティベートするiiidea  - target iiidea to deactivate.
    ///   - completion:
    ///     - isSuccess:ディアクティベートが成功した場合はtrue、失敗した場合はfalse
    ///     - error: RiiiverSDKから返却されたerror情報
    func deactivate(iiidea: ERApplet, completion: @escaping ((_ isSuccess: Bool, _ error: Error?) -> Void)) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        self.getAWSToken { (isSuccess, error) in
            if (isSuccess == false) {
                print("deactivate failed. reason: No token")
                completion(false, error)
                return
            }
            // Deactivate the iiidea via Riiiver SDK.
            iiidea.deactivate()
            print("deactivate succeeded.")
            completion(true, error)
        }
    }
    
    /// 指定したiiideaを削除する
    /// Remove the target iiidea.
    /// - Parameters:
    ///   - iiidea: 削除するiiidea  - target iiidea to remove.
    ///   - completion:
    ///     - isSuccess:削除が成功した場合はtrue、失敗した場合はfalse
    ///     - error: RiiiverSDKから返却されたerror情報
    func removeMyIiidea(iiidea: ERApplet, completion: @escaping ((_ isSuccess: Bool, _ error: Error?) -> Void)) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        self.getAWSToken { (isSuccess, error) in
            if (isSuccess == false) {
                print("removeMyIiidea failed. reason: No token")
                completion(false, error)
                return
            }
            
            ERSDK.sharedInstance.removeMyApplet(applet: iiidea) {(error) in
                guard error == nil else {
                    print("removeMyIiidea failed. reason: \(error.debugDescription))")
                    completion(false, error)
                    return
                }
                print("removeMyIiidea succeeded.")
                completion(true, error)
            }
        }
    }
    
    /// RiiiverSDKで生成したUUIDを取得する
    func getUuid() -> UUID? {
        let uuid = ERSDK.sharedInstance.getUuid()
        print("getUuid : \(String(describing: uuid))" )
        return uuid
    }

    /// 認証のパーミッション許可画面を表示する。
    /// Display a permission view to access an authentication.
    ///
    /// - Parameters:
    ///   - blockId: ID for target piece (block). - 対象ブロックのID。
    ///   - completion:
    ///     - isSuccess:削除が成功した場合はtrue、失敗した場合はfalse
    ///     - error: RiiiverSDKから返却されたerror情報
    func showAuthenticationSession(blockId:String, completion: @escaping ((_ isSuccess: Bool, _ error: Error?) -> Void)) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        self.getAWSToken { (isSuccess, error) in
            if (isSuccess == false) {
                print("showAuthenticationSession failed. reason: No token")
                completion(false, error)
                return
            }
            ERSDK.sharedInstance.getClientToken(completion: { (token, error) in
                guard error == nil else {
                    print("showAuthenticationSession failed. reason: \(error.debugDescription)")
                    completion(false, error)
                    return
                }
                // Auth認証用URLに必要な情報を設定してページを表示する
                // Configure the URL for O-Auth requirement and display the page.
                if let token: String = token, let uuid = RiiiverSDKManager.shared.getUuid()  {
                    let baseUrl = AppManager.riiiverAuthUri
                    let url = URL(string: baseUrl + "?blockId=\(blockId)&token=\(token)&uuid=\(uuid)")
                    UIApplication.shared.open(url!)
                    completion(true, nil)
                }
                else {
                    print("Web page display failure.")
                    completion(false, nil)
                }
            })
        }
    }
    
    /// 認証状態を解除する
    ///
    /// - Parameters:
    ///   - blockId: ID for target piece (block). - 対象ブロックのID。
    ///   - completion:
    ///     - isSuccess:削除が成功した場合はtrue、失敗した場合はfalse
    ///     - error: RiiiverSDKから返却されたerror情報
    func stopAuthenticationSession(blockId:String, completion: @escaping ((_ isSuccess: Bool, _ error: Error?) -> Void)) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        self.getAWSToken { (isSuccess, error) in
            if (isSuccess == false) {
                print("deleteClientAuth failed. reason: No token")
                completion(false, error)
                return
            }
            ERSDK.sharedInstance.deleteClientAuth(blockId: blockId) { error in
                guard error == nil else {
                    print("deleteClientAuth failed. reason: \(error.debugDescription)")
                    completion(false, error)
                    return
                }
                completion(true, error)
            }
        }
    }

    /// 指定したiiideaを強制的にダウンロードする
    /// Download the iiideas that your app wants to install in advance.  NOTE: Call this function before calling getMyApplets().
    /// - Parameters:
    ///   - appletIds: Target iiideas / 対象iiideaのID
    ///   - completion:
    ///     - isSuccess:削除が成功した場合はtrue、失敗した場合はfalse
    ///     - error: RiiiverSDKから返却されたerror情報
    func forcedDownloadApplet(completion: @escaping ((_ isSuccess: Bool, _ error: Error?) -> Void)) {
        // セッションが切れないように毎回トークンを更新  -  Update the token to keep the auth session.
        self.getAWSToken { (isSuccess, error) in
            if (isSuccess == false) {
                print("forcedDownloadApplet failed. reason: No token")
                completion(false, error)
                return
            }
            let appletIds:[String] = AppManager.shared.getDownloadAppletIds()
            self.forcedDownloadApplet(appletIds: appletIds) { isSuccess, error in
                guard error == nil else {
                    print("forcedDownloadApplet failed. reason: \(error.debugDescription)")
                    completion(false, error)
                    return
                }
                completion(true, error)
            }
        }
    }

    private func forcedDownloadApplet(appletIds:[String], completion: @escaping ((_ isSuccess: Bool, _ error: Error?) -> Void)) {
        if (appletIds.count == 0){
            print("No forced download applet")
            completion(true, nil)
            return
        }

        ERSDK.sharedInstance.forcedDownloadApplet(appletIds: appletIds) { error in
            guard error == nil else {
                print("forcedDownloadApplet failed. reason: \(error.debugDescription)")
                completion(false, error)
                return
            }
            completion(true, error)
        }
    }

    /// Riiiverからのお知らせを表示する
    /// - Parameters:
    ///   - webView: お知らせページを表示するWebView
    func showNews(webView: WKWebView) {
        ERNewsManager.sharedInstance.openNews(webView: webView, deviceIds: AppManager.shared.getDeviceId())
    }
    
    /// Riiiverからのお知らせを表示する
    /// - Parameters:
    ///   - webView: お知らせページを表示するWebView
    func isUnreadNews(completion: @escaping ((_ isUnread: Bool) -> Void)) {
        ERNewsManager.sharedInstance.checkNewsRead { isUpdated in
            completion(isUpdated)
        }
    }

    /// iiideaが実行開始した時に通知する
    /// - Parameters:
    ///   - applet: 起動したiiideaのインスタンス
    ///   - error: RiiiverSDKから返却されたerror情報
    func startApplet(applet: ERApplet, sdkError: ERSDKError) {
        print("startApplet : appletId=\(applet.getId()), appletTitle=\(applet.applet.getTitle()), sdkError=\(sdkError)" )
    }

    /// Activate状態が更新された時に通知する
    ///
    /// - ERSDK.getMyApplet() の呼び出し後、更新のあった iiidea を返却するコールバック
    /// - 更新前にActivate状態だったiiideaについては、iiidea更新後に再Activateを行う。
    /// - 再Activateに失敗したiiideaについては、 deactivatedToError で iiidea の ID とエラー内容のMapの形でコールバックされる
    /// - Parameters:
    ///   - updatedList: 更新されたPieceのリスト
    ///   - deactivatedToError: RiiiverSDKから返却されたerror情報
    func didUpdateAppletList(updatedList: [ERApplet], deactivatedToError: [String : ERSDKError]) {
        print("didUpdateAppletList : updatedList=\(updatedList), deactivatedToError=\(deactivatedToError)" )
    }

    /// iiideaが実行終了した時に通知する
    /// - Parameters:
    ///   - applet: 終了したiiideaのインスタンス
    ///   - error: iiideaが正しく終了しなかった時のエラー
    ///   - reachBlock: iiideatが終了した時の最終到達Block
    ///   - output: iiideaが正しく終了した際の、最後に到達したPieceが出力したデータ
    func endApplet(applet: ERApplet, sdkError: ERSDKError, reachBlock: ReachBlock, output: Dictionary<String, Any>?) {
        print("endApplet : appletId=\(applet.getId()), appletTitle=\(applet.applet.getTitle()), sdkError=\(sdkError), reachBlock=\(reachBlock), output=\(String(describing: output))")
    }
}
